#include <poll.h>      /* For poll()   */
#include <stdio.h>     /* For perror() */
#include <unistd.h>    /* For sleep()  */
#include <errno.h>     /* For errno    */

#include <vector>

using std::vector;

#include "headtask.hh"
#include "protocols.hh"
#include "socket.hh"
#include "stdin.hh"

typedef vector <struct HeadTask> HeadTaskListType;

class Head
{
    typedef vector <struct Socket>   SocketListType;

    typedef SocketListType::const_iterator SockCit;
    typedef SocketListType::iterator       SockIt;
    
    SocketListType Sockets;

    StdinType Stdin;
        
    void AddHeadTask(const HeadTask &task)
    {
        switch(task.Protocol)
        {
            case SockProtocolTCP:
            {
                struct Socket sock;
                sock.BecomeListenerTCP(task.ListenPort);
                Sockets.push_back(sock);
                break;
            }
            case SockProtocolUDP:
            {
                struct Socket sock;
                sock.BecomeListenerUDP(task.ListenPort);
                Sockets.push_back(sock);
                break;
            }
        }
    }
    
    struct pollfd AddPoll(const Socket &sock)
    {
        struct pollfd tmp = { sock.fd, POLLIN, 0 };
        
        if(sock.SendQueue.GetSize() > 0)
        {
            /* We want a chance to speak to the socket */
            tmp.events |= POLLOUT;
        }
        
        return tmp;
    }
    
    void MasterDied()
    {
        for(SockIt i=Sockets.begin(); i!=Sockets.end(); ++i)
            i->Die();
        
        Stdin.Die();
    }
    
    void HandleStdinPacket(const DataPacket &pak)
    {
        switch(pak.Type)
        {
            case DataPacket::PacketTypeNothing:
                break;
            
            case DataPacket::PacketTypeListen:
            {
                HeadTask tmp(pak.Peer.DestPort, pak.Protocol);
                AddHeadTask(tmp);
                break;
            }
            case DataPacket::PacketTypeData:
            {
                switch(pak.Protocol)
                {
                    case SockProtocolTCP:
                    {
                        /* Find the TCP connection */
                        SockIt i;
                        for(i=Sockets.begin(); i!=Sockets.end(); ++i)
                            if(i->Peer == pak.Peer)
                                break;
                        if(i != Sockets.end()) /* FIXME: What else? */
                        {
                            /* Found connection. Send the data there. */
                            i->SendTCP(pak.Data);
                        }
                        break;
                    }
                    case SockProtocolUDP:
                    {
                        /* This is much more complex. */
                        /* The UDP sockets are only listener sockets!
                         * There's no match for the Peer of this packet.
                         */
                        SockIt i;
                        for(i=Sockets.begin(); i!=Sockets.end(); ++i)
                            if(i->Peer.DestPort == pak.Peer.DestPort)
                                break;
                        if(i != Sockets.end()) /* FIXME: What else? */
                        {
                            /* Found the listener. Now make Peer match the recipient. */
                            i->Peer = pak.Peer;
                            
                            i->SendUDP(pak.Data);
                        }
                        break;
                    }
                }
                break;
            }
            case DataPacket::PacketTypeClose:
            {
                switch(pak.Protocol)
                {
                    case SockProtocolTCP:
                    {
                        /* Close a TCP connection. */
                        
                        SockIt i;
                        for(i=Sockets.begin(); i!=Sockets.end(); ++i)
                            if(i->Peer == pak.Peer)
                            {
                                /* Close a connection. */
                                
                                DataPacket pak = i->Die();
                                /* This will send a message again to the master
                                 * due to a deletion from connectioncache.
                                 * TODO: Make sure it does not matter.
                                 */
                                Stdin.SendPacket(pak);
                                Sockets.erase(i);
                                
                                /* No more connections to close. */
                                break;
                            }
                        break;
                    }
                    case SockProtocolUDP:
                    {
                        /* Bounce the packet. UDP is not connected,
                         * so it is therefore never closed too.
                         * They are simply forgetted.
                         */
                        Stdin.SendPacket(pak);
                        break;
                    }
                }
                break;
            }
            default:
                /* Weird stuff! */
                break;
        }
    }
    
    void HandlePollStdin(const struct pollfd &tmp)
    {
        if(tmp.revents & (POLLNVAL | POLLHUP))
        {
            /* Master does not want to hear us anymore! */
            
            MasterDied();

            return;
        }
        
        if(tmp.revents & POLLIN)
        {
            /* Master wants to say something! */
            
            DataPacket pak = Stdin.Read(STDIN_FILENO);
            
            HandleStdinPacket(pak);
        }
    }
    
    void HandlePollStdout(const struct pollfd &tmp)
    {
        if(tmp.revents & (POLLNVAL | POLLHUP))
        {
            /* Master does not want to talk to us anymore! */
            
            MasterDied();
        }
        if(tmp.revents & POLLOUT)
        {
            /* We may now speak to the master! */
            
            Stdin.Speak(STDOUT_FILENO);
        }
    }
    
    void HandlePollSocket(const struct pollfd &tmp, SockIt i)
    {
        Socket &sock = *i;
        
        if(tmp.revents & (POLLNVAL | POLLHUP))
        {
            /* This socket has died! */
            
            DataPacket pak = sock.Die();
            Stdin.SendPacket(pak);
            
            Sockets.erase(i);
            
            return;
        }
        
        switch(sock.Type)
        {
            case Socket::SockTypeListening:
            {
                switch(sock.Protocol)
                {
                    case SockProtocolTCP:
                    {
                        if(tmp.revents & POLLIN)
                        {
                            Socket newsock;
                            
                            /* `Fwiffo! Fwiffo darling! Would you
                             *  please answer the door? I think
                             *  someone's there.'
                             */
                            DataPacket pak = newsock.AcceptTCP(sock);
                            Stdin.SendPacket(pak);
                            
                            Sockets.push_back(newsock);
                            /* Invalidates sock (pointer) */
                        }
                        break;
                    }
                    case SockProtocolUDP:
                    {
                        if(tmp.revents & POLLIN)
                        {
                            /* UDP are connectionless sockets.
                             * You can directly try reading.
                             */
                            
                            DataPacket pak = sock.ReceiveUDP();
                            Stdin.SendPacket(pak);
                        }
                        if(tmp.revents & POLLOUT)
                        {
                            sock.SendUDP();
                        }
                        break;
                    }
                }
                break;
            }
            case Socket::SockTypeConnected:
            {
                switch(sock.Protocol)
                {
                    case SockProtocolTCP:
                    {
                        if(tmp.revents & POLLIN)
                        {
                            DataPacket pak = sock.ReceiveTCP();
                            Stdin.SendPacket(pak);
                        }
                        if(tmp.revents & POLLOUT)
                        {
                            sock.SendTCP();
                        }
                        break;
                    }
                    case SockProtocolUDP:
                    {
                        /* Empty: UDP sockets are never connected. */
                        /* This case should never hit. */
                        break;
                    }
                }
                break;
            }
        } /* switch(sock.type) */
    }
    
    void HandlePoll(const struct pollfd &tmp)
    {
        /* Is this stdin or stdout? */
        
        if(tmp.fd == STDIN_FILENO)
        {
            HandlePollStdin(tmp);
        }
        else if(tmp.fd == STDOUT_FILENO)
        {
            HandlePollStdout(tmp);
        }
        else
        {
            /* Find which socket this is */
            for(SockIt i=Sockets.begin(); i!=Sockets.end(); ++i)
                if(i->fd == tmp.fd)
                {
                    HandlePollSocket(tmp, i);
                    break;
                }
        }
    }
    
public:
    Head(const HeadTaskListType &Tasks)
    {
        /* Build the necessary sockets */
        
        HeadTaskListType::const_iterator i;
        for(i=Tasks.begin(); i!=Tasks.end(); ++i)
            AddHeadTask(*i);
        
        /* Main loop. It runs until stdin or stdout has been closed. */
        while(!Stdin.IsDead())
        {
            /* Add all sockets to the poll struct */
            vector<struct pollfd> polls;
            
            for(SockCit i=Sockets.begin(); i!=Sockets.end(); ++i)
                polls.push_back(AddPoll(*i));
            
            struct pollfd stdinfd = { STDIN_FILENO, POLLIN, 0 };
            struct pollfd stdoutfd = { STDOUT_FILENO, 0, 0 };

            if(Stdin.WantSpeak() > 0)
            {
                /* We want a chance to speak to stdout */
                stdoutfd.events |= POLLOUT;
            }
            
            /* Add stdin and stdout to the poll struct */
            
            polls.push_back(stdinfd);
            polls.push_back(stdoutfd);
            
            if(poll(&polls[0], polls.size(), -1) < 0)
            {
                if(errno == EINTR)
                {
                    /* Interrupted system call */
                    continue;
                }
                perror("poll");
                /* FIXME: And now what? */
                sleep(1);
                continue;
            }
            
            for(unsigned a=0; a<polls.size(); ++a)
                HandlePoll(polls[a]);
        }

        MasterDied();
    }
};


int main(void)
{
    HeadTaskListType HeadTasks;

    HeadTasks.push_back(HeadTask(6667, SockProtocolTCP));
    HeadTasks.push_back(HeadTask(4665, SockProtocolUDP));
    
    Head Run(HeadTasks);
    
    return 0;
}
